/* ebc356wd.c copyright 2000 by Crusader Systems (Pty) Ltd www.crusader.co.za
 * author: Charl P. Botha <cpbotha@ieee.org>
 * $Id: ebc356wd.c,v 1.5 2000/03/15 13:16:47 cpbotha Exp $
 *
 * linux driver for NEXCOM EBC356[P] built-in watchdog
 * http://www.nexcom.com/2-product/ebc/ebc356p.html
 *
 * thanks go to Crusader Systems (Pty) Ltd who have graciously allowed 
 * publishing under GPL
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 * Charl P. Botha <cpbotha@ieee.org> http://cpbotha.net
 * 
 * strange point 1: the cfg port for this watchdog is at 0xf2
 *                  in /usr/src/linux/arch/i386/kernel the region starting at
 *                  0xf0 and ending at 0xff is reserved for FPU!
 */

#include <linux/config.h>
#include <linux/module.h>
#include <linux/version.h>
#include <linux/types.h>
#include <linux/errno.h>
#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/miscdevice.h>
#include <linux/watchdog.h>
#include <linux/malloc.h>
#include <linux/ioport.h>
#include <linux/fcntl.h>
#include <asm/io.h>
#include <asm/uaccess.h>
#include <asm/system.h>
#include <linux/notifier.h>
#include <linux/reboot.h>
#include <linux/init.h>


#define WDIO 0xf2
#define WDSQW 0x70

static int ebc356wd_is_open=0;

static int ebc356wd_open(struct inode *inode, struct file *file) {
   if (MINOR(inode->i_rdev)!=WATCHDOG_MINOR) return -ENODEV;
   /* make sure of exclusive access */
   if (ebc356wd_is_open) return -EBUSY;
   MOD_INC_USE_COUNT;
   ebc356wd_is_open=1;
   /* first we initialize the SQW (square wave, what?) of the RTC */
   outb_p(0xa, WDSQW);
   outb_p(0x2f, WDSQW+1);
   outb_p(0xb, WDSQW);
   outb_p(0xa, WDSQW+1);
   /* clear the timer port (to make sure its initial value is zero */
   inb_p(WDIO);
   /* 7 6 5 4 3 2 1 0
    *       | | | | +----\
    *       | | | +------|--> 110==1s,101==2s,100==4s,011==8s,010==16s,001==32s,000==64s
    *       | | +--------/
    *       | +-------------> 1==enable wd, 0==disable wd
    *       +---------------> 1==select NMI, 0==select RESET
    * 
    * bits 7, 6 and 5 are don't cares
    *
    * Here we choose 8 seconds timeout, and out it to the configuration port.
    */
   outb_p(11,WDIO);
   
   return 0;
}

static int ebc356wd_release(struct inode *inode, struct file *file) {
   if (MINOR(inode->i_rdev)==WATCHDOG_MINOR) {
#ifndef CONFIG_WATCHDOG_NOWAYOUT
      /* clear the counter */
      inb_p(WDIO);
      /* disable the watchdog */
      outb_p(0, WDIO);
#endif
      ebc356wd_is_open=0;
   }
   MOD_DEC_USE_COUNT;
   return 0;
}

static ssize_t ebc356wd_write(struct file *file, const char *buf, size_t count, loff_t *ppos) {
   if (ppos!=&file->f_pos)
     return -ESPIPE;
   if (count) {
      /* clear the counter so we don't go boom */
      inb_p(WDIO);
      return 1;
   }
   return 0;
}

static int ebc356wd_notify_sys(struct notifier_block *this, unsigned long code,
	void *unused) {
	if(code==SYS_DOWN || code==SYS_HALT) {
	   /* zero the counter */
	   inb_p(WDIO);
	   /* turn the watchdog off */
	   outb_p(0, WDIO);
	}
	return NOTIFY_DONE;
}
 

static struct file_operations ebc356wd_fops = {
   NULL,
     NULL,
     ebc356wd_write,
     NULL,
     NULL,
     NULL,
     NULL,
     ebc356wd_open,
     NULL,
     ebc356wd_release
};
     
static struct miscdevice ebc356wd_miscdev = {
   WATCHDOG_MINOR,
     "watchdog",
     &ebc356wd_fops
};

static struct notifier_block ebc356wd_notifier = {
   ebc356wd_notify_sys,
     NULL,
     0
};

#ifdef MODULE
#define ebc356wd_init init_module

void cleanup_module(void) {
   misc_deregister(&ebc356wd_miscdev);
   unregister_reboot_notifier(&ebc356wd_notifier);
   release_region(WDIO, 1);
   release_region(WDSQW, 2);
}

#endif

__initfunc(int ebc356wd_init(void)) {
   printk("EBC356[P] watchdog driver $Revision: 1.5 $: <cpbotha@ieee.org>\n");
   misc_register(&ebc356wd_miscdev);
   request_region(WDIO, 1, "ebc356wd");
   request_region(WDSQW, 2, "ebc356wd");
   register_reboot_notifier(&ebc356wd_notifier);
   return 0;
}
